from __future__ import annotations
from typing import Any, Callable, Optional, Union

import collections
import contextlib
import functools
import math
import numpy as np
import os
import pathlib
import random
import torch


__all__ = [
    "adjacency",
    "batchwise_edge_mean",
    "batchwise_edge_sum",
    "batchwise_mask",
    "batchwise_node_mean",
    "batchwise_node_sum",
    "dot_product_attention",
    "expand_path",
    "in_degree",
    "int_length",
    "masked_softmax",
    "mkdir_safe",
    "memoize",
    "nodewise_edge_mean",
    "nodewise_edge_sum",
    "nodewise_mask",
    "out_degree",
    "set_seed",
    "temporary_seed",
]

Device = Union[str, torch.device]
Func = Callable[..., Any]


def adjacency(
    edge_index: torch.Tensor, device: Optional[Device] = "cpu"
) -> torch.Tensor:
    """Convert edge indices to adjacency matrix.

    Parameters
    ----------
    edge_index
        Edge index tensor.
        Shape: :math:`(2,N_{edges})`

    Returns
    -------
    torch.Tensor
        Adjacency matrix.
        Shape: :math:`(N_{nodes},N_{nodes})`
    """
    n = edge_index.max() + 1
    A = torch.zeros((n, n), device=device).long()
    A[tuple(edge_index)] = 1
    return A


# def aggregate_edges(
#     edges: torch.Tensor,
#     edge_index: torch.Tensor,
#     batch: torch.Tensor,
#     weights: Optional[torch.Tensor] = None,
# ) -> torch.Tensor:
#     if weights is None:
#         if nodewise:
#             weights = nodewise_mask(edge_index, device=edges.device)
#         else:
#             weights = batchwise_mask(batch, edge_index, device=edges.device)

#         if mean:
#             weights = torch.nn.functional.normalize(weights, p=1, dim=1)

#     return weights @ edges


def batchwise_edge_mean(
    edges: torch.Tensor,
    edge_index: torch.Tensor,
    batch: torch.Tensor,
    device="cpu",
) -> torch.Tensor:
    M = batchwise_mask(batch, edge_index, device=device)
    M = torch.nn.functional.normalize(M, p=1, dim=1)

    return M @ edges


def batchwise_edge_sum(
    edges: torch.Tensor,
    edge_index: torch.Tensor,
    batch: torch.Tensor,
    device="cpu",
) -> torch.Tensor:
    M = batchwise_mask(batch, edge_index, device=device)

    return M @ edges


def batchwise_mask(
    batch: torch.Tensor,
    edge_index: Optional[torch.Tensor] = None,
    device: Optional[Device] = "cpu",
) -> torch.Tensor:
    """Create a mask for batchwise aggregation of graph nodes or edges.

    Parameters
    ----------
    batch
        Tensor of batch indices.
        Shape: :math:`(N_{nodes},)`
    edge_index
        Tensor of edge indices, by default None.
        Shape: :math:`(2,N_{edges})`

    Returns
    -------
    torch.Tensor
        Mask for batchwise aggregation.
        Shape: :math:`(N_{batch},N_{nodes})` if `edge_index = None`
    """
    if edge_index is not None:
        # masking for edge aggregation
        _, N_e = edge_index.shape
        s, r = edge_index
        M = torch.zeros(batch[r].max() + 1, N_e, device=device)
        M[batch[r], torch.arange(N_e)] = 1
    else:
        # masking for node aggregation
        N_v = len(batch)
        M = torch.zeros(batch.max() + 1, N_v, device=device)
        M[batch, torch.arange(N_v)] = 1

    return M


def batchwise_node_mean(
    nodes: torch.Tensor, batch: torch.Tensor, device="cpu"
) -> torch.Tensor:
    M = batchwise_mask(batch, device=device)
    M = torch.nn.functional.normalize(M, p=1, dim=1)

    return M @ nodes


def batchwise_node_sum(
    nodes: torch.Tensor, batch: torch.Tensor, device="cpu"
) -> torch.Tensor:
    M = batchwise_mask(batch, device=device)

    return M @ nodes


def dot_product_attention(
    query: torch.Tensor, key: torch.Tensor, mask: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """Compute scaled dot product attention.

    Parameters
    ----------
    query
        Query vectors.
        Shape: :math:`(N_{queries},d_q)`
    key
        Key vectors.
        Shape: :math:`(N_{keys},d_q)`
    mask
        Mask tensor to ignore query-key pairs, by default None.
        Shape: :math:`(N_{queries},N_{keys})`

    Returns
    -------
    torch.Tensor
        Scaled dot product attention between each query and key vector.
        Shape: :math:`(N_{queries},N_{keys})`
    """

    _, d = key.shape

    pre_attn = query @ key.transpose(0, 1) / math.sqrt(d)

    return masked_softmax(pre_attn, mask, dim=1)


def expand_path(path: str, dir: Optional[str] = None) -> str:
    p = pathlib.Path(path).expanduser()
    if dir is not None:
        p = pathlib.Path(dir).joinpath(p)
    return str(p.expanduser().resolve())


def in_degree(adjacency: torch.Tensor) -> torch.Tensor:
    """Compute in-degrees of nodes in a graph.

    Parameters
    ----------
    adjacency
        Adjacency matrix.
        Shape: :math:`(N_{nodes}, N_{nodes})`

    Returns
    -------
    torch.Tensor
        Nodewise in-degree tensor.
        Shape: :math:`(N_{nodes},)`
    """
    return adjacency.sum(dim=0)


def int_length(n: int) -> int:
    # from https://stackoverflow.com/a/2189827
    if n > 0:
        return int(math.log10(n)) + 1
    elif n == 0:
        return 1
    else:
        # +1 if you don't count the '-'
        return int(math.log10(-n)) + 2


def masked_softmax(
    x: torch.Tensor, mask: torch.Tensor, dim: Optional[int] = 1
) -> torch.Tensor:
    """Compute softmax of a tensor using a mask.

    Parameters
    ----------
    x
        Argument of softmax.
    mask
        Mask tensor.
    dim
        Dimension along which softmax will be computed.

    Returns
    -------
    torch.Tensor
        Masked softmax of `x`.
    """
    X = (mask * x).masked_fill(mask == 0, -float("inf"))
    return torch.softmax(X, dim)


def mkdir_safe(directory: str) -> None:
    with contextlib.suppress(FileExistsError):
        os.makedirs(directory)


def memoize(func: Func) -> Func:
    func.cache = collections.OrderedDict()

    @functools.wraps(func)
    def memoized(*args, **kwargs):
        k = (args, frozenset(kwargs.items()))
        if k not in func.cache:
            func.cache[k] = func(*args, **kwargs)

        return func.cache[k]

    return memoized


def nodewise_edge_mean(
    edges: torch.Tensor, edge_index: torch.Tensor, device="cpu"
) -> torch.Tensor:
    M = nodewise_mask(edge_index, device=device)
    M = torch.nn.functional.normalize(M, p=1, dim=1)

    return M @ edges


def nodewise_edge_sum(
    edges: torch.Tensor, edge_index: torch.Tensor, device="cpu"
) -> torch.Tensor:
    """Sum all incoming edge features for each node.

    Parameters
    ----------
    edges
        Edge features.
        Shape: :math:`(N_{edges},d_e)`
    edge_index
        Ege indices.
        Shape: :math:`(2,N_{edges})`

    Returns
    -------
    torch.Tensor
        Tensor of summed incoming edges for each node.
        Shape: :math:`(N_{nodes},d_e)`
    """
    M = nodewise_mask(edge_index, device=device)

    return M @ edges


def nodewise_mask(edge_index: torch.Tensor, device="cpu") -> torch.Tensor:
    """Create a mask for nodewise aggregation of incoming graph edges.

    Parameters
    ----------
    edge_index
        Edge indices.
        Shape: :math:`(2,N_{edges})`

    Returns
    -------
    torch.Tensor
        Mask for nodewise edge aggregation.
        Shape: :math:`(N_{nodes},N_{edges})`
    """
    _, N_e = edge_index.shape
    s, r = edge_index
    M = torch.zeros(r.max() + 1, N_e, device=device)
    M[r, torch.arange(N_e)] = 1

    return M


def out_degree(adjacency: torch.Tensor) -> torch.Tensor:
    """Compute out-degrees of nodes in a graph.

    Parameters
    ----------
    adjacency
        Adjacency matrix.
        Shape: :math:`(N_{nodes},N_{nodes})`

    Returns
    -------
    torch.Tensor
        Nodewise out-degree tensor.
        Shape: :math:`(N_{nodes},)`
    """
    return adjacency.sum(dim=1)


def set_seed(seed: int) -> None:
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


@contextlib.contextmanager
def temporary_seed(seed: int) -> None:
    # adapted from https://stackoverflow.com/a/49557127
    random_state = random.getstate()
    np_state = np.random.get_state()
    torch_state = torch.random.get_rng_state()

    set_seed(seed)

    try:
        yield
    finally:
        random.setstate(random_state)
        np.random.set_state(np_state)
        torch.random.set_rng_state(torch_state)



from __future__ import annotations
from typing import Any, Callable, Iterable, Optional, Union

import luz
import torch

__all__ = ["Trainer"]


Device = Union[str, torch.device]
Loss = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]


class Trainer:
    def __init__(
        self,
        loss: Optional[Loss] = None,
        optimizer: Optional[luz.Optimizer] = None,
        start_epoch: Optional[int] = 1,
        stop_epoch: Optional[int] = 2,
        early_stopping: Optional[bool] = False,
        handlers: Optional[Iterable[luz.Handler]] = None,
        patience: Optional[int] = 5,
        **loader_kwargs: Union[int, bool, luz.Transform],
    ) -> None:
        """Algorithm to train a model using data.

        Parameters
        ----------
        loss
            Loss function to be minimized during training, by default None
        optimizer
            Training optimizer, by default None
        start_epoch
            First training epoch, by default 1
        stop_epoch
            Last training epoch, by default 2
        early_stopping
            If True, then use `val_dataset` for early stopping; by default False.
            Ignored if `val_dataset` is `None`.
        patience
            Number of epochs of non-improving validation loss
            before training stops early; by default 5.
            Ignored if `early_stopping` is `False`.
        handlers
            Handlers to run, by default None.
        **loader_kwargs
            Dataset.loader kwargs.
        """
        self.loss = loss
        self.optimizer = optimizer
        self.start_epoch = start_epoch
        self.stop_epoch = stop_epoch
        self.early_stopping = early_stopping
        self.patience = patience
        self.loader_kwargs = loader_kwargs
        self.handlers = handlers
        self._state = {}

    def state_dict(self):
        return {
            "loss": self.loss,
            "optimizer": self.optimizer,
            "start_epoch": self.start_epoch,
            "stop_epoch": self.stop_epoch,
            "early_stopping": self.early_stopping,
            "patience": self.patience,
            "loader_kwargs": self.loader_kwargs,
            "handlers": self.handlers,
            # "_state": self._state,
        }

    def load_state_dict(self, state_dict):
        self.loss = state_dict["loss"]
        self.optimizer = state_dict["optimizer"]
        self.start_epoch = state_dict["start_epoch"]
        self.stop_epoch = state_dict["stop_epoch"]
        self.early_stopping = state_dict["early_stopping"]
        self.patience = state_dict["patience"]
        self.loader_kwargs = state_dict["loader_kwargs"]
        self.handlers = state_dict["handlers"]
        # self._state = state_dict["_state"]

    def fit(
        self,
        model: luz.Module,
        dataset: luz.Dataset,
        val_dataset: Optional[luz.Dataset] = None,
        device: Optional[Device] = "cpu",
    ) -> None:
        """Train model.

        Parameters
        ----------
        model
            Model to be trained.
        dataset
            Training data.
        val_dataset
            Validation data, by default None.
        device
            Device to use for training, by default "cpu".
        """
        model.migrate(device)

        # NOTE: must come after migrate
        optimizer = self.optimizer.link(model=model)

        loader = dataset.loader(**self.loader_kwargs)
        self._state = {}
        self.log(
            flag=luz.Flag.TRAINING,
            trainer=self,
            model=model,
            optimizer=optimizer,
            loader=loader,
            train_history=[],
            val_history=[],
        )
        if self.early_stopping:
            self.log(patience=self.patience)

        self._call_event(luz.Event.TRAINING_STARTED)

        for epoch in range(self.start_epoch, self.stop_epoch):
            self.log(epoch=epoch)
            self._call_event(luz.Event.EPOCH_STARTED)

            self.run_epoch(model, loader, device, True, optimizer)

            if val_dataset is not None:
                self.log(flag=luz.Flag.VALIDATING)
                val_loss = self.validate(model, val_dataset, device)
                print(f"[Epoch {epoch}] Validation loss: {val_loss}.")
                if self.early_stopping and self._state["patience"] == 0:
                    print(f"[Epoch {epoch}]: Stopping early.")
                    break
                self.log(flag=luz.Flag.TRAINING)

        self._call_event(luz.Event.TRAINING_ENDED)

    def validate(
        self, model: luz.Module, dataset: luz.Dataset, device: Optional[Device] = "cpu"
    ) -> float:
        """Validate model.

        Parameters
        ----------
        model
            Model to be validated.
        dataset
            Validation data.
        device
            Device to use for validation, by default "cpu".

        Returns
        -------
        float
            Validation loss.
        """
        loader = dataset.loader(**self.loader_kwargs)
        with model.eval():
            val_loss = self.run_epoch(model, loader, device, train=False)

        self._state["val_history"].append(val_loss)

        try:
            # FIXME: replace 0.0 with self.delta_thresh?
            if min(self._state["val_history"]) - val_loss < 0.0:
                self._state["patience"] -= 1
            else:
                self._state["patience"] = self.patience
        except ValueError:
            pass

        return val_loss

    def test(
        self,
        model: luz.Module,
        dataset: luz.Dataset,
        device: Optional[Device] = "cpu",
    ) -> float:
        """Test model.

        Parameters
        ----------
        model
            Model to be tested.
        dataset
            Test data.
        device
            Device to use for testing, by default "cpu".
        """
        model.migrate(device)

        loader = dataset.loader(**self.loader_kwargs)
        self._state = {}
        self.log(
            flag=luz.Flag.TESTING,
            trainer=self,
            model=model,
            loader=loader,
        )
        self._call_event(luz.Event.TESTING_STARTED)

        self.log(epoch=1)
        self._call_event(luz.Event.EPOCH_STARTED)

        test_loss = self.run_epoch(model, loader, device, train=False)

        self._call_event(luz.Event.TESTING_ENDED)

        return test_loss

    def run_epoch(
        self,
        model: luz.Module,
        loader: torch.utils.data.DataLoader,
        device: Device,
        train: bool,
        optimizer: Optional[torch.optim.Optimizer] = None,
    ) -> float:
        running_loss = 0.0

        for i, batch in enumerate(loader):
            data = model.get_input(batch)
            target = model.get_target(batch)

            self.log(ind=i, data=data, target=target)
            self._call_event(luz.Event.BATCH_STARTED)

            # migrate the input and target tensors to the appropriate device
            data, target = data.to(device), target.to(device)

            if train:
                running_loss += model.run_batch(data, target, device, optimizer)
            else:
                with model.eval():
                    running_loss += model.run_batch(data, target, device)
            # from https://coolnesss.github.io/2019-02-05/pytorch-gotchas
            # All of the variables defined above are now out of scope!
            # On CPU, they are already deallocated.
            # On GPU, they will be deallocated soon.

            # Make sure deallocation has taken place
            if torch.cuda.is_available():
                torch.cuda.synchronize()

            self._call_event(luz.Event.BATCH_ENDED)

        loss = running_loss / len(loader)

        if train:
            self._state["train_history"].append(loss)

        self._call_event(luz.Event.EPOCH_ENDED)

        return loss

    def log(self, **kwargs: Any) -> None:
        self._state.update(**kwargs)

    def _call_event(self, event: luz.Event) -> None:
        for h in self.handlers:
            getattr(h, event.name.lower())(**self._state)



from __future__ import annotations
from typing import Any, Iterable, Optional

from abc import ABC, abstractmethod
import collections
import luz
import networkx
import numpy as np
import torch

__all__ = [
    "Argmax",
    "Center",
    "CenterPerTensor",
    "Compose",
    "DigraphToTensors",
    "Expand",
    "Identity",
    "Lookup",
    "Normalize",
    # "NormalizePerTensor",
    "PowerSeries",
    "Standardize",
    "Transform",
    "Transpose",
    "ZeroMeanPerTensor",
]


class Transform:
    def __init__(self, **transforms: TensorTransform) -> None:
        self.transforms = collections.defaultdict(Identity, transforms)

    def __call__(self, data: luz.Data) -> luz.Data:
        kw = {k: self.transforms[k](data[k]) for k in data.keys}
        return luz.Data(**kw)


class TensorTransform(ABC):
    @abstractmethod
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        "Transform which is applied to an input tensor."


class Argmax(TensorTransform):
    def __init__(
        self, dim: Optional[int] = None, keepdim: Optional[bool] = False
    ) -> None:
        self.dim = dim
        self.keepdim = keepdim

    def __call__(self, x: torch.Tensor) -> torch.LongTensor:
        return x.argmax(dim=self.dim, keepdim=self.keepdim)


class Center(TensorTransform):
    def __init__(self, dataset: luz.Dataset, key: str) -> None:
        for x in dataset:
            self.mean = torch.zeros_like(x[key])
            break

        n = len(dataset)

        for i in range(n):
            delta = dataset[i][key] - self.mean
            self.mean += delta / (i + 1)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x - self.mean


class CenterPerTensor(TensorTransform):
    def __init__(self, dataset: luz.Dataset, key: str, dim: int) -> None:
        for x in dataset:
            self.mean = torch.zeros_like(x[key].mean(dim))
            break

        n = len(dataset)

        for i in range(n):
            delta = dataset[i][key].mean(dim) - self.mean
            self.mean += delta / (i + 1)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x - self.mean


class Compose(TensorTransform):
    def __init__(self, *transforms: Iterable[TensorTransform]) -> None:
        self.transforms = transforms

    def __call__(self, x: Any) -> Any:
        for transform in self.transforms:
            x = transform(x)
        return x


class DigraphToTensors(TensorTransform):
    def __call__(self, x: networkx.DiGraph):
        # FIXME: replace np.vstack with torch.stack
        nodes = torch.Tensor(np.vstack([x.nodes[n]["x"] for n in x.nodes]))
        edge_index = torch.Tensor(list(x.edges)).long().t().contiguous()

        return nodes, edge_index


class Expand(TensorTransform):
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x.unsqueeze(-1)


class Identity(TensorTransform):
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x


class Lookup(TensorTransform):
    def __init__(self, lookup_dict) -> None:
        self.lookup_dict = lookup_dict

    def __call__(self, x) -> Any:
        return self.lookup_dict[x]


# class NormalizePerTensor(TensorTransform):
#     def __call__(self, x: torch.Tensor) -> torch.Tensor:
#         x -= torch.mean(x)
#         x /= torch.std(x)
#         return x


class Normalize(TensorTransform):
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x / x.abs().max()


class PowerSeries(TensorTransform):
    def __init__(self, degree: int, dim: Optional[int] = -1) -> None:
        self.degree = degree
        self.dim = dim

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        tensors = tuple(x ** k for k in range(1, self.degree + 1))
        try:
            return torch.cat(tensors=tensors, dim=self.dim)
        except RuntimeError:
            return torch.stack(tensors=tensors, dim=self.dim)


class Standardize(TensorTransform):
    def __init__(self, dataset: luz.Dataset, key: str) -> None:
        for x in dataset:
            self.mean = torch.zeros_like(x[key])
            variance = torch.zeros_like(x[key])
            break

        n = len(dataset)

        for i in range(n):
            delta = dataset[i][key] - self.mean
            self.mean += delta / (i + 1)
            variance += delta * (dataset[i][key] - self.mean)

        self.std = torch.sqrt(variance / n)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self.mean) / self.std


class Transpose(TensorTransform):
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x.t()


class ZeroMeanPerTensor(TensorTransform):
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return x - torch.mean(x)



from __future__ import annotations
from typing import Optional, Union

from abc import ABC, abstractmethod
import luz
import torch

__all__ = ["Learner"]

Device = Union[str, torch.device]


class Learner(ABC):
    @abstractmethod
    def learn(
        self,
        dataset: luz.Dataset,
        val_dataset: Optional[luz.Dataset] = None,
        device: Optional[Device] = "cpu",
    ) -> luz.Module:
        """Learn a model based on a given dataset.

        Parameters
        ----------
        dataset
            Training dataset used to learn a model.
        val_dataset
            Validation dataset, by default None.
        device
            Device to use for learning, by default "cpu".

        Returns
        -------
        luz.Module
            Learned model.
        """
        pass



"""

Contains callback objects which perform various functions during the training process.

"""

# FIXME: Write a test for every handler here (many are likely broken)
# FIXME: Type annotate every handler here

from __future__ import annotations
from typing import Any, Iterable, Optional, Union

import datetime
import luz
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import os
import pathlib
import torch

__all__ = [
    "Handler",
    "Accuracy",
    "ActualVsPredicted",
    # "Checkpoint",
    # "DataAndFit",
    # "DurbinWatson",
    "FBeta",
    # "LogToFile",
    "Loss",
    "PlotHistory",
    "Progress",
    # "RVP",
    "Timer",
]


class Handler:
    def batch_started(self, **kwargs: Any):
        pass

    def batch_ended(self, **kwargs: Any):
        pass

    def epoch_started(self, **kwargs: Any):
        pass

    def epoch_ended(self, **kwargs: Any):
        pass

    def testing_started(self, **kwargs: Any):
        pass

    def testing_ended(self, **kwargs: Any):
        pass

    def training_started(self, **kwargs: Any):
        pass

    def training_ended(self, **kwargs: Any):
        pass


class Accuracy(Handler):
    def __init__(self) -> None:
        self.correct = 0
        self.total = 0

    def epoch_started(self, **kwargs: Any) -> None:
        """Compute on epoch start."""
        self.correct = 0
        self.total = 0

    def batch_ended(
        self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any
    ) -> None:
        """Compute on batch end.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,C)`
        target
            Target tensor. One-hot encoded.
            Shape: :math:`(N,C)`
        """
        predicted = torch.argmax(torch.softmax(output, dim=1), dim=1)
        correct = torch.argmax(target, dim=1)

        self.correct += (predicted == correct).sum().item()
        self.total += target.size(0)

    def epoch_ended(self, epoch: int, **kwargs: Any) -> None:
        acc = self.correct / self.total
        s = f"[Epoch {epoch}] Classification accuracy: {acc}"
        print(s)


class ActualVsPredicted(Handler):
    def __init__(self, filepath: Optional[Union[str, pathlib.Path]] = None) -> None:
        """Plot actual labels vs. predicted labels.

        Parameters
        ----------
        filepath
            Path to save plot if not None, by default None.
        """
        self.filepath = filepath

    def batch_ended(
        self, output: torch.Tensor, target: torch.Tensor, flag: luz.Flag, **kwargs: Any
    ) -> None:
        """Compute on batch end.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,)`
        target
            Target tensor.
            Shape: :math:`(N,)`
        """
        if flag == luz.Flag.TESTING:
            x = output.detach().reshape(-1).numpy()
            y = target.detach().reshape(-1).numpy()
            self.ax.scatter(x, y, color="black")

    def testing_started(self, **kwargs: Any) -> None:
        """Compute on testing start."""
        self.fig, self.ax = plt.subplots()

    def testing_ended(self, **kwargs: Any) -> None:
        """Compute on testing end."""
        x_min, x_max = self.ax.get_xlim()
        y_min, y_max = self.ax.get_ylim()
        plot_min = min(x_min, y_min)
        plot_max = max(x_max, y_max)
        self.ax.set_xlim(plot_min, plot_max)
        line = mlines.Line2D([0, 1], [0, 1], color="red")
        line.set_transform(self.ax.transAxes)
        self.ax.add_line(line)
        plt.ylabel("Predicted")
        plt.xlabel("Actual")
        plt.title("Predicted vs. actual")
        if self.filepath is not None:
            plt.savefig(luz.expand_path(self.filepath))

        plt.show()


class Checkpoint(Handler):
    def __init__(self, save_interval, model_name, save_dir=None):
        if save_interval < 0:
            raise ValueError(
                "The number of epochs between checkpoints must be a positive integer."
            )

        self.model_name = model_name
        self.save_dir = luz.utils.expand_path(path=save_dir)

        luz.mkdir_safe(self.save_dir)

        self.save_interval = save_interval

    # # FIXME: Fix this
    # def state_dict(self):
    #     return {}

    def epoch_ended(self, model, epoch, **kwargs):
        if epoch % self.save_interval == 0:
            save_path = os.path.join(
                self.save_dir, f"{model.model_name}_{epoch}.pth.tar"
            )
            torch.save(obj=model.state_dict(), f=save_path)


class DataAndFit(Handler):
    def __init__(self):
        self.xs = []
        self.actual = []
        self.predicted = []

    def update(self, x, y, predicted):
        self.xs.append(x.item())
        self.actual.append(y.item())
        self.predicted.append(predicted.item())

    def compute(self):
        plt.scatter(self.xs, self.actual, label="Actual data")
        plt.scatter(self.xs, self.predicted, label="Fit function")
        plt.xlabel("Input")
        plt.ylabel("Output")
        plt.title("Data and fit")
        plt.show()


class DurbinWatson(Handler):
    def __init__(self):
        self.numerator = 0
        self.denominator = 0
        self.last_residual = None

    def batch_ended(self, x, y, output, **kwargs):
        residual = (y - output).item()
        if self.last_residual is not None:
            self.numerator += (residual - self.last_residual) ** 2
        self.last_residual = residual
        self.denominator += residual ** 2

    def compute(self):
        print("DW: {0}".format(self.numerator / self.denominator))
        return self.numerator / self.denominator


class FBeta(Handler):
    def __init__(self, beta: float) -> None:
        self.beta = beta

        self.true_positive = 0
        self.predicted_positive = 0
        self.actual_positive = 0

    def epoch_started(self, **kwargs: Any) -> None:
        """Compute on epoch start."""
        self.true_positive = 0
        self.predicted_positive = 0
        self.actual_positive = 0

    def batch_ended(
        self, output: torch.Tensor, target: torch.Tensor, **kwargs: Any
    ) -> None:
        """Compute on batch end.

        Parameters
        ----------
        output
            Output tensor.
            Shape: :math:`(N,2)`
        target
            Target tensor. One-hot encoded.
            Shape: :math:`(N,2)`
        """
        predicted = torch.argmax(torch.softmax(output, dim=1), dim=1)
        correct = torch.argmax(target, dim=1)

        self.true_positive += correct[predicted.nonzero(as_tuple=False)].sum().item()
        self.predicted_positive += predicted.sum().item()
        self.actual_positive += correct.sum().item()

    def epoch_ended(self, epoch: int, **kwargs: Any) -> None:
        try:
            precision = self.true_positive / self.predicted_positive
            recall = self.true_positive / self.actual_positive
            F = (
                (1 + self.beta ** 2)
                * precision
                * recall
                / (precision * (self.beta ** 2) + recall)
            )
        except ZeroDivisionError:
            F = 1

        s = f"[Epoch {epoch}] F-score: {F}"
        print(s)


class LogToFile(Handler):
    def __init__(self, log_path, log_list):
        # FIXME: rewrite this whole class
        assert False, "TODO: Finish writing LogToFile"

        self.log_path = log_path
        self.log_list = log_list

    # def _compute_confusion_matrix(self):

    def on_epoch_end(self):
        with open(self.log_path, "w") as f:
            for log_compute in self.log_list:
                f.write(log_compute)


class Loss(Handler):
    def __init__(self, print_interval: Optional[float] = 0.25) -> None:
        """Calculate running loss throughout each epoch.

        Parameters
        ----------
        print_interval
            Fraction of epoch after which the running loss is printed, by default 0.25.

        Raises
        ------
        ValueError
        """
        if print_interval <= 0 or print_interval > 1:
            raise ValueError(
                "Print interval must be a positive number between 0.0 and 1.0."
            )

        self.print_interval = print_interval

    def epoch_started(self, **kwargs: Any) -> None:
        self.running_loss = 0

    def batch_ended(
        self,
        flag: luz.Flag,
        loader: torch.utils.data.DataLoader,
        epoch: int,
        loss: torch.Tensor,
        ind: int,
        **kwargs: Any,
    ) -> None:
        """Execute at end of batch.

        Parameters
        ----------
        flag
            Training flag.
        loader
            Data loader.
        epoch
            Epoch number.
        loss
            Last computed loss.
        ind
            Data index.
        **kwargs
            Superfluous kwargs.
        """
        if flag == luz.Flag.TRAINING:
            # NOTE: it's very important to add loss.item()
            # (as opposed to loss) to avoid a memory leak!
            self.running_loss += loss.item()
            num_batches = len(loader)
            batch_interval = max(1, round(num_batches * self.print_interval))
            cur = ind + 1
            if cur % batch_interval == 0 or cur == num_batches:
                avg_loss = self.running_loss / cur / loader.batch_size
                print(f"[Epoch {epoch}] Running average loss: {avg_loss}.")


class PlotHistory(Handler):
    def training_ended(
        self,
        train_history: Iterable[float],
        val_history: Iterable[float],
        **kwargs: Any,
    ) -> None:
        x = range(1, len(train_history) + 1)
        fig, ax1 = plt.subplots()
        ax1.set_xlabel("Epoch")
        (line1,) = ax1.plot(x, train_history, color="tab:blue")

        if len(val_history) > 0:
            ax2 = ax1.twinx()
            (line2,) = ax2.plot(x, val_history, color="tab:orange")

            plt.title("Loss history")
            plt.legend((line1, line2), ("Training loss", "Validation loss"))
        else:
            plt.title("Loss history")
            plt.legend((line1,), ("Training loss",))

        fig.tight_layout()
        plt.show()


class Progress(Handler):
    def __init__(
        self, print_interval: Optional[float] = 0.25, bar_length: Optional[int] = 30
    ) -> None:
        if print_interval <= 0 or print_interval > 1:
            raise ValueError(
                "Print interval must be a positive number between 0.0 and 1.0."
            )

        self.bar_length = bar_length
        self.print_interval = print_interval

    def batch_ended(
        self,
        loader: torch.utils.data.DataLoader,
        epoch: int,
        flag: luz.Flag,
        ind: int,
        **kwargs: Any,
    ) -> None:
        if flag == luz.Flag.TRAINING:
            progress = (ind + 1) / len(loader)

            if int(progress / self.print_interval) > self._last or progress == 1:
                num_eqs = int(progress * self.bar_length)
                num_spaces = self.bar_length - num_eqs
                bar = "=" * num_eqs + ">" + " " * num_spaces
                print(f"[Epoch {epoch}]: [{bar}] {int(100*progress)}%")
                self._last += 1

    def epoch_started(self, **kwargs: Any) -> None:
        self._last = 0

    def training_started(self, **kwargs: Any) -> None:
        print("Training started.")

    def training_ended(self, **kwargs: Any) -> None:
        print("Training ended.")

    def testing_started(self, flag, **kwargs: Any) -> None:
        print("Testing started.")

    def testing_ended(self, flag, **kwargs: Any) -> None:
        print("Testing complete.")


class RVP(Handler):
    def __init__(self) -> None:
        self.predicted = []
        self.residual = []

    def batch_ended(
        self, x: torch.Tensor, y: torch.Tensor, output: torch.Tensor, **kwargs: Any
    ) -> None:
        self.predicted.append(output.item())
        self.residual.append((y - output).item())

    def training_ended(self, **kwargs: Any) -> None:
        self.fig, self.ax = plt.subplots()
        self.ax.scatter(self.predicted, self.residual, c="black")
        mmax = max([abs(max(self.residual)), abs(min(self.residual))])
        plt.ylim([-mmax, mmax])
        plt.axhline(y=0.0, color="r", linestyle="-")
        plt.ylabel("Residual")
        plt.xlabel("Predicted")
        plt.title("Residual vs. predicted")
        plt.show()


class Timer(Handler):
    def __init__(self) -> None:
        self.start_time = 0
        self.end_time = 0

    def epoch_started(self, **kwargs: Any) -> None:
        self.start_time = datetime.datetime.now().replace(microsecond=0)

    def epoch_ended(self, epoch: int, flag: luz.Flag, **kwargs: Any) -> None:
        self.end_time = datetime.datetime.now().replace(microsecond=0)

        if flag == luz.Flag.TRAINING:
            mode = "train"
        elif flag == luz.Flag.VALIDATING:
            mode = "validate"
        else:
            mode = "test"

        print(f"[Epoch {epoch}]: {self.end_time-self.start_time} to {mode}.")
